/*! ********************************************
 * VoiceUnitFM2.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth 
{
internal final class VoiceUnitFM2 extends VoiceUnitBase
{
// Tables
	// Wave
	private const _WAVE_TABLE_MASK :int = SynthesizerFM2.WAVE_TABLE_SIZE - 1;
	// PG
	private const _PG_PHASE_FP      :int = SynthesizerFM2.PG_PHASE_FP;
	private const _PG_PHASE_FP_MASK :int = (1 << _PG_PHASE_FP) - 1;  // after the fixed-point
	private const _PG_PHASE_MASK    :int = (SynthesizerFM2.WAVE_TABLE_SIZE << _PG_PHASE_FP) - 1;
	// FM
	private var _MUL_TABLE :Vector.<int> = null;
	// LFO
	private var   _LFO_WAVE_TABLES      :Vector.<Vector.<int>> = null;
	private const _LFO_TIMER_INIT_VALUE :int = SynthData.LFO_TIMER_INIT_VALUE;
	private const _LFO_WAVE_TABLE_MASK  :int = SynthData.LFO_WAVE_TABLE_SIZE - 1;
	// Synth
	internal var VOICE_PARTS     :Vector.<VoicePart> = null;
	internal var WAVE_TABLES     :Vector.<Vector.<int>> = null;
	internal var TONE_PRESETS    :Vector.<TonePresetFM> = null;
	internal var SAMPLE_DATAS    :Vector.<SampleData> = null;
	internal var SAMPLER_PRESETS :Vector.<SamplerPreset> = null;
	internal var pgSamplePow2tIdx :int = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
	internal var pgKeyPow2tIdx    :int = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
	internal var egPow2tIdx       :int = 0;  // EG PhaseStep (Pow2Table Index Offset)
	internal var egPhaseStepSndOff:int = 0;  // EG PhaseStep - SoundOff
	internal var lfoTimerCoef     :int = 0;  // Coefficient Value - LFO Phase Update Timer
	// Pitch
	private var _ptBend     :int = 0;   // pitchBend
	private var _ptBendSens :int = 0;   // pitchBend Sensitivity
	// LFO
	private var _lfoPhase     :int = 0;  // LFO Phase
	private var _lfoTimer     :int = 0;  // LFO Phase Update Timer
	private var _lfoTimerStep :int = 0;  // Update Timer Step
	private var _lfoMod_0     :int = 0;  // Modulation Depth
	private var _lfoMod_1     :int = 0;  // Modulation Depth
	private var _lfoModSens_0 :int = 0;  // Modulation Sensitivity
	private var _lfoModSens_1 :int = 0;  // Modulation Sensitivity
	private var _lfoOutPM_0   :int = 0;  // output - phase modulation
	private var _lfoOutPM_1   :int = 0;  // output - phase modulation
	private var _lfoOutAM_0   :int = 0;  // output - amplitude modulation
	private var _lfoOutAM_1   :int = 0;  // output - amplitude modulation
// Operator 0,1 ---
	private var _lfoWaveTable :Vector.<int> = null;
	private var _pgBend       :int = 0;  // v = _ptBend * _ptBendSens
	private var _fdbk         :int = 0;
	
// Operator 0 ---
	// Wave
	private var _waveTable_0  :Vector.<int> = null;
	// PG
	private var _pgKey_0      :int = 0;
	private var _pgPhase_0    :int = 0;
	private var _pgPhaseStep_0:int = 0;
	// EG
	/*
	- Envelop Generator State -
	[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
	[ ---- RSDA ] "-" = Reserved Bits
		A = Attack  stage
		D = Decay   stage
		S = Sustain stage
		R = Release stage
	*/
	private var _egState_0      :int = 0;
	private var _egLevelTable_0 :Vector.<int> = null;
	private var _egCount_0      :int = 0;
	private var _egCountStep_0  :int = 0;
	//private var _egCountStepAR:int = 0;
	private var _egCountStepDR_0:int = 0;
	private var _egCountStepRR_0:int = 0;
	private var _egCntEnvSus_0  :int = 0;
	private var _egCntEnvBtm_0  :int = 0;
	private var _egGate_0       :int = 0;
	private var _egSL_0         :int = 0;
	private var _egTL_0         :int = 0;
	private var _egTLv_0        :int = 0;
// Operator 1 ---
	// Wave
	private var _waveTable_1  :Vector.<int> = null;
	// PG
	private var _pgKey_1      :int = 0;
	private var _pgPhase_1    :int = 0;
	private var _pgPhaseStep_1:int = 0;
	// EG
	private var _egState_1      :int = 0;
	private var _egLevelTable_1 :Vector.<int> = null;
	private var _egCount_1      :int = 0;
	private var _egCountStep_1  :int = 0;
	//private var _egCountStepAR_1:int = 0;
	private var _egCountStepDR_1:int = 0;
	private var _egCountStepRR_1:int = 0;
	private var _egCntEnvSus_1  :int = 0;
	private var _egCntEnvBtm_1  :int = 0;
	private var _egGate_1       :int = 0;
	private var _egSL_1         :int = 0;
	private var _egTL_1         :int = 0;
	private var _egTLv_1        :int = 0;
// Sampler
	private var _pgPhaseAFP :int = 0;  // PG Phase (After the Fixed Point)
	private var _plbkE      :int = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
	private var _loopE      :int = 0;  // Loop End Frame       [loopE <= plbkE]
	private var _loopF      :int = 0;  // Loop Frames          [val = loopE - lpStart]
	private var _loopFk     :int = 0;  // Loop Frames          * after -> keyOff
	
	override internal function keyOff():void
	{
		// KeyOn -->
		if (state & 0x0010)
		{
		// OP0 is Active -->
			if (state & 1)
			{
				if (_egState_0 == 1)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
					else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_0 = _egCntEnvBtm_0;
				_egCountStep_0 = _egCountStepRR_0;
				_egState_0 = 8;
				if (_egCount_0 > _egCntEnvBtm_0)
				{ 
					if (state & 2){ state &= 0xf2; _egState_0 = 0; }
					else { killProcess(); return; }
				}
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
		// OP1 is Active -->
			if (state & 2)
			{
				if (_egState_1 == 1)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
					else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_1 = _egCntEnvBtm_1;
				_egCountStep_1 = _egCountStepRR_1;
				_egState_1 = 8;
				if (_egCount_1 > _egCntEnvBtm_1)
				{ 
					if (state & 1){ state &= 0xf1; _egState_1 = 0; }
					else { killProcess(); return; }
				}
			}
			state = 0x0020 | state & 3;
		}
	}
	override internal function soundOff():void
	{
		// KeyOn || KeyOff --->
		if (state & 0x0030)
		{
		// OP0 Active --->
			if (state & 1)
			{
				if (_egState_0 == 1)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
					else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_0 > _egCntEnvBtm_0)
				{ 
					if (state & 2){ state &= 0xf2; }
					else { killProcess(); return; }
				}
				_egGate_0 = _egCntEnvBtm_0;
				_egCountStep_0 = egPhaseStepSndOff;
				_egState_0 = 0;
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
		// OP1 Active --->
			if (state & 2)
			{
				if (_egState_1 == 1)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
					else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_1 > _egCntEnvBtm_1)
				{ 
					if (state & 1){ state &= 0xf1; }
					else { killProcess(); return; }
				}
				_egGate_1 = _egCntEnvBtm_1;
				_egCountStep_1 = egPhaseStepSndOff;
				_egState_1 = 0;
			}
			state = 0x0040 | state & 3;
			partNumber = -1; // SoundOff
		}
	}
	private function _shiftEGState_0():int
	{
		switch (_egState_0)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				_egCount_0 = 0;
				// -> DECAY
				if (_egSL_0 > 0)
				{
					_egGate_0 = _egCntEnvSus_0;
					_egCountStep_0 = _egCountStepDR_0;
					return _egState_0 = 2;
				}
				// -> SUSTAIN
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				return _egState_0 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_0 + _egSL_0) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					_egGate_0 = 0x7fffffff;
					_egCountStep_0 = 0;
					_egCount_0 = _egCntEnvSus_0;
					return _egState_0 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_0 = 0;
		}
	}
	private function _shiftEGState_1():int
	{
		switch (_egState_1)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				_egCount_1 = 0;
				// -> DECAY
				if (_egSL_1 > 0)
				{
					_egGate_1 =  _egCntEnvSus_1;
					_egCountStep_1 = _egCountStepDR_1;
					return _egState_1 = 2;
				}
				// -> SUSTAIN
				_egGate_1 = 0x7fffffff;
				_egCountStep_1 = 0;
				return _egState_1 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_1 + _egSL_1) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					_egGate_1 = 0x7fffffff;
					_egCountStep_1 = 0;
					_egCount_1 = _egCntEnvSus_1;
					return _egState_1 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_1 = 0;
		}
	}
	override internal function keyOn(ch:int, note:int, velo:int):void
	{
	// KeyOn Status ---
		state = 0x0013;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		var _part:VoicePart = VOICE_PARTS[partNumber], _preset:TonePresetFM, 
		_keyFlw_0:int = egPow2tIdx, 
		_keyFlw_1:int = egPow2tIdx;
		if (_part.mode == 2)
		{
			_preset = TONE_PRESETS[noteNumber | 128];
			_pgKey_0 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_0] + (_preset.tune_0 - 8192) + pgKeyPow2tIdx;
			_pgKey_1 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_1] + (_preset.tune_1 - 8192) + pgKeyPow2tIdx;
			//_keyFlw = 0;
			_prPan = _preset.pan;
		}
		else
		{
			_preset = TONE_PRESETS[_part.programChange];
			_pgKey_0 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_0] + (_preset.tune_0 - 8192) + pgKeyPow2tIdx;
			_pgKey_1 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_1] + (_preset.tune_1 - 8192) + pgKeyPow2tIdx;
			_keyFlw_0 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_0] << _preset.kf_0;
			_keyFlw_1 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_1] << _preset.kf_1;
			_prPan = 64;
		}
	// LFO ---
		_lfoWaveTable = _LFO_WAVE_TABLES[_preset.lfoWave];
		_lfoModSens_0 = _preset.lms_0;
		_lfoModSens_1 = _preset.lms_1;
		_lfoMod_0 = _part.ccValues[1] * _lfoModSens_0;
		_lfoMod_1 = _part.ccValues[1] * _lfoModSens_1;
		_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
		_lfoTimer = _LFO_TIMER_INIT_VALUE;
		_lfoPhase = _preset.strc & 1;
		_lfoOutPM_0 = 0;
		_lfoOutPM_1 = 0;
		_lfoOutAM_0 = 0;
		_lfoOutAM_1 = 0;
	// Velocity Control ---
		_egTL_0 = _preset.tl_0 + (127 - velo >> 7 - _preset.tlVelCtrl_0);
		if (_egTL_0 > 127)_egTL_0 = 0;
		_egTL_1 = _preset.tl_1 + (127 - velo >> 7 - _preset.tlVelCtrl_1);
		if (_egTL_1 > 127)_egTL_1 = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// pitch ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
	// process ---
		process = _PROC_PTRS[_preset.strc];
		_fdbk = _EG_TOTAL_LEVEL_TABLE[_preset.fl];
// Operator 0 --->
		_waveTable_0 = WAVE_TABLES[_preset.wave_0];
	// PG ---
		var v:int = _pgKey_0 + _pgBend;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_0 = 0;
	// EG ---
		_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
		_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
		_egSL_0 = _preset.sl_0;
		_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
		if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
		v = (_preset.dr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
		_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
		_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_0 < 127) {
			_egCount_0 = _EG_CNT_ENV_ATK;
			_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
			_egGate_0 = _EG_MAX_CNT;
			v = (_preset.ar_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
			_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_0 = 1;
		} else {
			_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
			_egCount_0 = 0;
		// -> DECAY
			if (_egSL_0 > 0) {
				_egGate_0 = _egCntEnvSus_0;
				_egCountStep_0 = _egCountStepDR_0;
				_egState_0 = 2;
		// -> SUSTAIN
			} else {
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				_egState_0 = 4; }
		}
// Operator 1 --->
		_waveTable_1 = WAVE_TABLES[_preset.wave_1];
	// PG ---
		v = _pgKey_1 + _pgBend;
		_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_1 = 0;
	// EG ---
		_egTLv_1 = _EG_TOTAL_LEVEL_TABLE[_egTL_1];
		_egCntEnvBtm_1 = _EG_CNT_GATE_TABLE[128 - _egTL_1];
		_egSL_1 = _preset.sl_1;
		_egCntEnvSus_1 = _EG_CNT_GATE_TABLE[_egSL_1];
		if (_egCntEnvSus_1 > _egCntEnvBtm_1)_egCntEnvSus_1 = _egCntEnvBtm_1;
		v = (_preset.dr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
		_egCountStepDR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
		_egCountStepRR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_1 < 127) {
			_egCount_1 = _EG_CNT_ENV_ATK;
			_egLevelTable_1 = _EG_LEVEL_TABLE_UP;
			_egGate_1 = _EG_MAX_CNT;
			v = (_preset.ar_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
			_egCountStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_1 = 1;
		} else {
			_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
			_egCount_1 = 0;
		// -> DECAY
			if (_egSL_1 > 0) {
				_egGate_1 = _egCntEnvSus_1;
				_egCountStep_1 = _egCountStepDR_1;
				_egState_1 = 2;
		// -> SUSTAIN
			} else {
				_egGate_1 = 0x7fffffff;
				_egCountStep_1 = 0;
				_egState_1 = 4; }
		}
	}
	internal function keyOnSampler(ch:int, note:int, velo:int):void
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		var _part:VoicePart = VOICE_PARTS[partNumber],
			_preset:SamplerPreset = SAMPLER_PRESETS[noteNumber],
			_sample:SampleData = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable_0 = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoOutPM_0 = 0;
		_lfoOutAM_0 = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = _preset.pan;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey_0 = (_preset.octv - 8 << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
		var v:int = _pgKey_0 + _pgBend;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_0 = _sample.plbkS;
		_pgPhaseAFP = 0;
		_pgKey_1 = 0; _pgPhase_1 = 0;
	// EG ---
		_egTL_0 = _preset.tl;
		_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
		_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
		_egSL_0 = _preset.sl;
		_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
		if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount_0 = _EG_CNT_ENV_ATK;
			_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
			_egGate_0 = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
			_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_0 = 1;
		} else {
			_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
			_egCount_0 = 0;
		// -> DECAY
			if (_egSL_0 > 0) {
				_egGate_0 = _egCntEnvSus_0;
				_egCountStep_0 = _egCountStepDR_0;
				_egState_0 = 2;
		// -> SUSTAIN
			} else {
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				_egState_0 = 4; }
		}
	// process ---
		process = (_sample.channelCount == 2)? _procSmpl2ch: _procSmpl1ch;
	}
	override internal function updPartBendSens(val:int):void
	{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		var v:int = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		// Inst op2 --->
		if (state & 2)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
	}
	override internal function updPartPitchBend(val:int):void
	{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		var v:int = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		// Inst op2 --->
		if (state & 2)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
	}
	override internal function updPartModulation(val:int):void
	{
		_lfoMod_0 = val * _lfoModSens_0;
		_lfoMod_1 = val * _lfoModSens_1;
	}
	internal function updateLfoFrequency(val:int):void
	{
		_lfoTimerStep = val * lfoTimerCoef;
	}
// Constructor
	public function VoiceUnitFM2():void
	{
		super();
		var _synthData:SynthData = SynthData.getInstance();
		// Log2 Volume Table
		_LOG2V_TABLE    = _synthData.getLOG2V_TABLE();
		_LOG2V_TABLE_I  = _synthData.getLOG2V_TABLE_I();
		// Pow2 Table
		_POW2_TABLE      = _synthData.getPOW2_TABLE();
		_POW2T_RS_TABLE  = _synthData.getPOW2T_RS_TABLE();
		_POW2T_KEY_TABLE = _synthData.getPOW2T_KEY_TABLE();
		// EG
		_EG_TOTAL_LEVEL_TABLE = _synthData.getEG_TOTAL_LEVEL_TABLE();
		_EG_LEVEL_TABLE_UP    = _synthData.getEG_LEVEL_TABLE_UP();
		_EG_LEVEL_TABLE_DW    = _synthData.getEG_LEVEL_TABLE_DW();
		_EG_KEY_TABLE         = _synthData.getEG_KEY_TABLE();
		_EG_POW2T_RS_TABLE    = _synthData.getEG_POW2T_RS_TABLE();
		_EG_CNT_GATE_TABLE    = _synthData.getEG_CNT_GATE_TABLE();
		// etc
		_MUL_TABLE       = _synthData.getMUL_TABLE();
		_LFO_WAVE_TABLES = _synthData.getLFO_WAVE_TABLES();
		
		_PROC_PTRS[0] = _proc0;
		_PROC_PTRS[1] = _proc1;
		_PROC_PTRS[2] = _proc2;
		_PROC_PTRS[3] = _proc3;
		_PROC_PTRS[4] = _proc4;
		_PROC_PTRS[5] = _proc5;
		_PROC_PTRS[6] = _proc6;
		_PROC_PTRS[7] = _proc7;
	}
// process ---
	// pointer array (process Function)
	private const _PROC_PTRS :Vector.<Function> = new Vector.<Function>(8, true);
	// [0] -> out
	// [1] -> out
	private function _proc0(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				x = _wt_0[_pgPhase_0 >> _PG_PHASE_FP] + x;
			// OP 0 out --->
				n = _log2vT[x];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else
					{
						if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				n += _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		}
		startFrame = i;
	}
	// [0 (FeedBack)] -> out
	// [1           ] -> out
	private function _proc1(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				n = _log2vT[p];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else
					{
						if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				n += _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		}
		startFrame = i;
	}
	// [LFO(PM)][0 (FeedBack)] -> out
	//          [1           ] -> out
	private function _proc2(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				// op0
				_lfoOutPM_0 = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				n = _log2vT[p];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else
					{
						if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				n += _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	// [LFO(AM)][0 (FeedBack)] -> out
	//          [1           ] -> out
	private function _proc3(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				x = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				n = _log2vT[p];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else
					{
						if (state & 1){ state &= 0xf1; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				n += _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	// [0] -> [1] -> out
	private function _proc4(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				x = _wt_0[_pgPhase_0 >> _PG_PHASE_FP] + x;
			// OP 0 out --->
				p = _log2vTI[x];
			// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else p = 0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else { killProcess(); return; }
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			} else n = 0.0;
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		}
		startFrame = i;
	}
	// [0 (FeedBack)] -> [1] -> out
	private function _proc5(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				p = _log2vTI[p];
			// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else p = 0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else { killProcess(); return; }
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			} else n = 0.0;
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		}
		startFrame = i;
	}
	// [LFO(PM)][0(FeedBack)] -> [1] -> out
	private function _proc6(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				// op0
				_lfoOutPM_0 = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				p = _log2vTI[p];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else p = 0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else { killProcess(); return; }
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			} else n = 0.0;
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	// [LFO(AM)][0(FeedBack)] -> [1] -> out
	private function _proc7(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, x:int, p:int, f:int,
			_et_0:Vector.<int> = _egLevelTable_0, _et_1:Vector.<int> = _egLevelTable_1;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, _log2vTI:Vector.<int> = _LOG2V_TABLE_I,
			_wt_0:Vector.<int> = _waveTable_0, _wt_1:Vector.<int> = _waveTable_1;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				x = (_lfoWaveTable[_lfoPhase] * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (_lfoWaveTable[_lfoPhase] * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if (state & 1)
			{
			// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if (_shiftEGState_0())_et_0 = _egLevelTable_0;
					else
					{
						if (state & 2){ state &= 0xf2; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				p = _log2vTI[p];
			// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else p = 0;
		// Operator 1 --->
			if (state & 2)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if (_shiftEGState_1())_et_1 = _egLevelTable_1;
					else { killProcess(); return; }
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				p = (p + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			} else n = 0.0;
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
// process --- Perc
	// Sampler 1ch
	private function _procSmpl1ch(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, v:int,
			_et:Vector.<int> = _egLevelTable_0;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable_0;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et = _egLevelTable_0;
				else { killProcess(); return; }
			}
			v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
		// pg update --->
			if (_pgPhase_0 >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase_0] + v;
			n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhaseAFP += _pgPhaseStep_0;
			_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 2ch
	private function _procSmpl2ch(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, v:int, p:int, 
			_et:Vector.<int> = _egLevelTable_0;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable_0;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0())_et = _egLevelTable_0;
				else { killProcess(); return; }
			}
			v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
		// pg update --->
			if (_pgPhase_0 >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			p = _pgPhase_0 << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhaseAFP += _pgPhaseStep_0;
			_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
}
}